@@ -13,8 +13,14 @@
 ##############################################################################
 """ExceptionFormatter tests.
 """
-import unittest
+import importlib
+import pathlib
 import sys
+import tempfile
+import textwrap
+import unittest
+import zipfile
+from urllib.error import HTTPError
 
 
 IS_PY39_OR_GREATER = sys.version_info >= (3, 9)
@@ -157,13 +163,11 @@ class TextExceptionFormatterTests(unittest.TestCase):
                          '   - __traceback_info__: XYZZY')
 
     def test_formatTracebackInfo_unicode(self):
-        __traceback_info__ = u"Have a Snowman: \u2603"
+        __traceback_info__ = "Have a Snowman: \u2603"
         fmt = self._makeOne()
 
         result = fmt.formatTracebackInfo(__traceback_info__)
-        expected = '   - __traceback_info__: Have a Snowman: '
-        # utf-8 encoded on Python 2, unicode on Python 3
-        expected += '\xe2\x98\x83' if bytes is str else u'\u2603'
+        expected = '   - __traceback_info__: Have a Snowman: \u2603'
         self.assertIsInstance(result, str)
         self.assertEqual(result, expected)
 
@@ -278,6 +282,37 @@ class TextExceptionFormatterTests(unittest.TestCase):
                          ''.join(
                              traceback.format_exception_only(ValueError, err)))
 
+    def test_formatExceptionOnly_httperror(self):
+        # On Python 3.11.0 HTTPError may behave wrongly, giving a KeyError in
+        # tempfile when Python tries to format it.
+        # See https://github.com/python/cpython/issues/90113
+        # or examples in Plone tests, especially doctests:
+        # https://github.com/plone/Products.CMFPlone/issues/3663
+        # We don't want to get an error when we format an error,
+        # so let's compensate in our code.
+        fmt = self._makeOne()
+        err = HTTPError('url', 400, 'oops', [], None)
+        result = fmt.formatExceptionOnly(HTTPError, err).strip()
+        # The output can differ too much per Python version,
+        # but it is just one line when stripped.
+        self.assertIn("400", result)
+        self.assertIn("oops", result)
+        self.assertIn("Error", result)
+        self.assertEqual(len(result.splitlines()), 1)
+
+    def test_formatException_httperror(self):
+        # See test_formatExceptionOnly_httperror.
+        # Here we check that formatException works.
+        fmt = self._makeOne()
+        err = HTTPError('url', 400, 'oops', [], None)
+        result = fmt.formatException(HTTPError, err, None)
+        self.assertEqual(result[0], 'Traceback (most recent call last):\n')
+        last = result[-1]
+        # The output can differ per Python version.
+        self.assertIn("400", last)
+        self.assertIn("oops", last)
+        self.assertIn("Error", last)
+
     def test_formatLastLine(self):
         fmt = self._makeOne()
         self.assertEqual(fmt.formatLastLine('XXX'), 'XXX')
@@ -484,6 +519,10 @@ class HTMLExceptionFormatterTests(unittest.TestCase):
         fmt = self._makeOne()
         self.assertEqual(fmt.escape('XXX'), 'XXX')
 
+    def test_escape_non_str(self):
+        fmt = self._makeOne()
+        self.assertEqual(fmt.escape(123), '123')
+
     def test_escape_w_markup(self):
         fmt = self._makeOne()
         self.assertEqual(fmt.escape('<span>XXX & YYY<span>'),
@@ -650,7 +689,7 @@ class Test_format_exception(unittest.TestCase):
             self.assertTrue(s.find('level%d' % n) >= 0, s)
 
     def test_quote_last_line(self):
-        class C(object):
+        class C:
             pass
         try:
             raise TypeError(C())
@@ -671,7 +710,7 @@ class Test_format_exception(unittest.TestCase):
                         lines[2])
 
     def test_traceback_info_non_ascii(self):
-        __traceback_info__ = u"Have a Snowman: \u2603"
+        __traceback_info__ = "Have a Snowman: \u2603"
         try:
             raise TypeError()
         except TypeError:
@@ -709,14 +748,14 @@ class Test_format_exception(unittest.TestCase):
 
     def test_format_exception_as_html(self):
         # Test for format_exception (as_html=True)
-        from zope.exceptions.exceptionformatter import format_exception
-        from textwrap import dedent
         import re
+
+        from zope.exceptions.exceptionformatter import format_exception
         try:
             exec('import')
         except SyntaxError:
             result = ''.join(format_exception(*sys.exc_info(), as_html=True))
-        expected = dedent("""\
+        expected = textwrap.dedent("""\
             <p>Traceback (most recent call last):</p>
             <ul>
             <li>  Module {module}, line ABC, in {fn}<br />
@@ -739,6 +778,32 @@ class Test_format_exception(unittest.TestCase):
         self.maxDiff = None
         self.assertEqual(expected, result)
 
+    def test_pep302_loader_source_in_traceback(self):
+        sys_path = sys.path[::]
+
+        def restore_sys_path():
+            sys.path[::] = sys_path
+        self.addCleanup(restore_sys_path)
+
+        module_name = self._testMethodName
+        with tempfile.TemporaryDirectory() as td:
+            td = pathlib.Path(td)
+            module_zipfile = td / f"{module_name}.zip"
+            with zipfile.ZipFile(module_zipfile, "w") as zf:
+                zf.writestr(
+                    zipfile.ZipInfo(f"{module_name}/__init__.py"),
+                    textwrap.dedent("""
+                        def f():
+                            1 / 0
+                    """))
+            sys.path.insert(0, str(module_zipfile))
+            try:
+                importlib.import_module(module_name).f()
+            except ZeroDivisionError:
+                s = self._callFUT(False)
+
+        self.assertIn('1 / 0', s)
+
 
 class Test_print_exception(unittest.TestCase):
 
@@ -811,7 +876,7 @@ class Test_extract_stack(unittest.TestCase):
 
     def test_traceback_info_html(self):
         try:
-            __traceback_info__ = u"Adam & Eve"
+            __traceback_info__ = "Adam & Eve"
             raise ExceptionForTesting
         except ExceptionForTesting:
             s = self._callFUT(True)
@@ -820,7 +885,7 @@ class Test_extract_stack(unittest.TestCase):
     def test_traceback_supplement_text(self):
         try:
             __traceback_supplement__ = (TestingTracebackSupplement,
-                                        u"You're one in a million")
+                                        "You're one in a million")
             raise ExceptionForTesting
         except ExceptionForTesting:
             s = self._callFUT(False)
@@ -866,7 +931,7 @@ class ExceptionForTesting(Exception):
     pass
 
 
-class TestingTracebackSupplement(object):
+class TestingTracebackSupplement:
     source_url = '/somepath'
     line = 634
     column = 57
@@ -876,7 +941,7 @@ class TestingTracebackSupplement(object):
         self.expression = expression
 
 
-class DummySupplement(object):
+class DummySupplement:
     def __init__(self, info=''):
         self._info = info
 
@@ -884,12 +949,15 @@ class DummySupplement(object):
         return self._info
 
 
-class DummyTB(object):
+class DummyTB:
+    # https://docs.python.org/3/reference/datamodel.html#traceback-objects
+    tb_frame = None
     tb_lineno = 14
     tb_next = None
+    tb_lasti = 1
 
 
-class DummyFrame(object):
+class DummyFrame:
     f_lineno = 137
     f_back = None
 
@@ -899,12 +967,23 @@ class DummyFrame(object):
         self.f_code = DummyCode()
 
 
-class DummyCode(object):
+class DummyCode:
     co_filename = 'dummy/filename.py'
     co_name = 'dummy_function'
 
+    def co_positions(self):
+        # New in Python 3.11.
+        # https://docs.python.org/3/reference/datamodel.html#codeobject.co_positions
+        # Note that this is not called for DummyTB if you have tb_lasti=-1.
+        # The 27 in the return value is chosen to match tb_recurse.tb_lineno=27
+        # in test_formatException_recursion_in_tb_stack in this file.
+        # The rest is random.
+        # Note that this code is only called on Python 3.11+, so we mark it for
+        # the coverage tool.
+        return [(27, 2, 3, 4)]  # pragma: no cover
 
-class _Monkey(object):
+
+class _Monkey:
     # context-manager for replacing module names in the scope of a test.
     def __init__(self, module, **kw):
         self.module = module
@@ -922,7 +1001,3 @@ class _Monkey(object):
                 setattr(self.module, key, value)
             else:
                 delattr(self.module, key)
-
-
-def test_suite():
-    return unittest.defaultTestLoader.loadTestsFromName(__name__)
